import 'dart:async';
import 'dart:convert';

import 'package:flutter/services.dart';
import 'package:flutter_socket_io/iflutter_socket_io.dart';

String _parserFunctionName(Function function) {
  if (function != null) {
    String funcName = function.toString();

    //Closure: (dynamic) => dynamic from Function '_socketStatus@16413367'
    if (funcName != null &&
        funcName.isNotEmpty &&
        "'".allMatches(funcName).length > 1) {

      funcName = funcName.substring(funcName.indexOf("'") + 1, funcName.lastIndexOf("'"));
      if(funcName != null && funcName.isNotEmpty) {
        print("funcName: " + funcName);
        return funcName;
      }
    }
  }
  return null;
}

class SocketIOFunction {
  String _funcName;
  Function _function;

  SocketIOFunction(Function function) {
    _function = function;
    _funcName = _parserFunctionName(function);
  }

  set setFunction(Function function) {
    _function = function;
    _funcName = _parserFunctionName(function);
  }

  String get functionName => _funcName;

  Function get function => _function;
}

class CallbackFunctions {

  List<SocketIOFunction> _functions;

  CallbackFunctions() {
    _functions = new List();
  }

  SocketIOFunction addFunction(Function function) {
    SocketIOFunction f = getFunction(function);
    if (f == null) {
      f = new SocketIOFunction(function);
      _functions.add(f);
    }
    return f;
  }

  SocketIOFunction getFunctionByName(String funcName) {
    if (funcName != null && funcName.isNotEmpty) {
      SocketIOFunction result;
      _functions.forEach((f) {
        if (f != null && f.functionName == funcName) {
          result = f;
          return;
        }
      });
      return result;
    }
    return null;
  }

  SocketIOFunction getFunction(Function function) {
    if (function != null) {
      SocketIOFunction result;
      String funcName = _parserFunctionName(function);
      _functions.forEach((f) {
        if (f != null && funcName != null && f.functionName == funcName) {
          result = f;
          return;
        }
      });
      return result;
    }
    return null;
  }

  bool remove(Function function) {
    if (function != null) {
      String funcName = _parserFunctionName(function);
      SocketIOFunction socketIOFunction;
      _functions.forEach((f) {
        if (f != null && funcName != null && f.functionName == funcName) {
          socketIOFunction = f;
          return;
        }
      });

      return _functions.remove(socketIOFunction);
    }
    return false;
  }

  void removeAllFunctions() {
    if (_functions != null) {
      _functions.clear();
    }
  }

  List<SocketIOFunction> get functions => _functions;
}

class SocketIO implements ISocketIO {

  static const MethodChannel _channel = const MethodChannel('flutter_socket_io');

  String _domain;
  String _namespace;
  Function _statusCallback;
  String _statusCallbackName;

  Map<String, CallbackFunctions> _callbacks;

  ///domain: domain url
  /// namespace: just for iOS
  /// socketStatusCallback [optional]: the status of socket will be sent into this function
  SocketIO(String domain, String namespace, {Function socketStatusCallback}) {
    _domain = domain;
    _namespace = namespace;
    _statusCallback = socketStatusCallback;
    _statusCallbackName = _parserFunctionName(socketStatusCallback);
    _channel.setMethodCallHandler(handlerMethodCall);
    _callbacks = new Map();
  }

  dynamic _parserMethodCall(String method) {
    if (method != null && method.isNotEmpty && method.contains("|")) {
      return method.split("|");
    }
  }

  Future<dynamic> handlerMethodCall(MethodCall call) {
    print("method: " + call.method);

    //invoke function String with structure: event|method
    //example: socket_info|onSocketInfo
    if (call.method.contains("|")) {

      if(call.method.contains("_onReceiveChatMessage")) {
        print(call.method);
      }

      dynamic params = _parserMethodCall(call.method);
      print(params);

      if (params != null && params.length > 1) {
        String event = params[0];
        if (event != null && event.isNotEmpty) {

          CallbackFunctions functions = _callbacks[event];
          if (functions != null) {
            SocketIOFunction f = functions.getFunctionByName(params[1]);
            if (f != null && f.function != null) {
              print("CALLLING FUNCTION: " + f.functionName);
              f.function(call.arguments);
            }
          }

        }
      }
    } else if (call.method == _statusCallbackName && _statusCallback != null) {
      _statusCallback(call.arguments);
    }
  }

  _clearAll() {
    if(_callbacks != null) {
      _callbacks.clear();
    }
  }

  @override
  void connect() async {
    await _channel.invokeMethod(MethodCallName.SOCKET_CONNECT, {
      MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
      MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace,
      MethodCallArgumentsName.SOCKET_CALLBACK: _statusCallbackName
    });
  }

  @override
  void subscribe(String event, Function callback) async {
    if (event != null && event.isNotEmpty) {
      CallbackFunctions functions = _callbacks[event];
      SocketIOFunction f;

      if (functions == null) {
        functions = new CallbackFunctions();
      }

      if (callback != null) {
        f = functions.addFunction(callback);
      }

      _callbacks[event] = functions;

      var subscribes = new Map<String, String>();
      subscribes.putIfAbsent(event, () => f == null ? "" : f.functionName);

      await _channel.invokeMethod(MethodCallName.SOCKET_SUBSCRIBES, {
        MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
        MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace,
        MethodCallArgumentsName.SOCKET_DATA: json.encode(subscribes),
      });
    }
  }

  @override
  void unSubscribe(String event, Function callback) async {

    if (event != null && event.isNotEmpty) {

      CallbackFunctions callbackFunctions = _callbacks[event];

      if (callbackFunctions != null) {

        callbackFunctions.remove(callback);

        if(callbackFunctions.functions.length < 1) {
          _callbacks.remove(event);
        } else {
          _callbacks[event] = callbackFunctions;
        }

      }

      SocketIOFunction f = new SocketIOFunction(callback);
      var unSubscribes = new Map<String, String>();
      unSubscribes.putIfAbsent(event, () => f == null ? "" : f.functionName);
      await _channel.invokeMethod(MethodCallName.SOCKET_UNSUBSCRIBES, {
        MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
        MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace,
        MethodCallArgumentsName.SOCKET_DATA: json.encode(unSubscribes),
      });
    }
  }

  @override
  void sendMessage(String event, dynamic message, [Function callback]) async {
    if (event != null && event.isNotEmpty) {
      CallbackFunctions functions = _callbacks[event];
      SocketIOFunction f;

      if (functions == null) {
        functions = new CallbackFunctions();
      }

      if (callback != null) {
        f = functions.addFunction(callback);
      }

      _callbacks[event] = functions;

      await _channel.invokeMethod(MethodCallName.SOCKET_SEND_MESSAGE, {
        MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
        MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace,
        MethodCallArgumentsName.SOCKET_EVENT: event,
        MethodCallArgumentsName.SOCKET_MESSAGE: message,
        MethodCallArgumentsName.SOCKET_CALLBACK: f == null ? "" : f.functionName
      });
    }
  }

  @override
  void disconnect() async {
    await _channel.invokeMethod(MethodCallName.SOCKET_DISCONNECT, {
      MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
      MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace,
    });
  }

  @override
  String getId() {
    if (_domain != null) {
      return _domain + (_namespace != null ? _namespace : "");
    }
    return null;
  }

  @override
  void destroy() async {
    _clearAll();
    await _channel.invokeMethod(MethodCallName.SOCKET_DESTROY, {
      MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
      MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace
    });
  }

  @override
  void unSubscribesAll() async {
    _clearAll();
    await _channel.invokeMethod(MethodCallName.SOCKET_UNSUBSCRIBES_ALL, {
      MethodCallArgumentsName.SOCKET_DOMAIN: _domain,
      MethodCallArgumentsName.SOCKET_NAME_SPACE: _namespace
    });
  }
}

class MethodCallArgumentsName {
  static final String SOCKET_DOMAIN = "socketDomain";
  static final String SOCKET_NAME_SPACE = "socketNameSpace";
  static final String SOCKET_CALLBACK = "socketCallback";
  static final String SOCKET_EVENT = "socketEvent";
  static final String SOCKET_MESSAGE = "socketMessage";
  static final String SOCKET_DATA = "socketData";
}

class MethodCallName {
  static final String SOCKET_CONNECT = "socketConnect";
  static final String SOCKET_DISCONNECT = "socketDisconnect";
  static final String SOCKET_SUBSCRIBES = "socketSubcribes";
  static final String SOCKET_UNSUBSCRIBES = "socketUnsubcribes";
  static final String SOCKET_UNSUBSCRIBES_ALL = "socketUnsubcribesAll";
  static final String SOCKET_SEND_MESSAGE = "socketSendMessage";
  static final String SOCKET_DESTROY = "socketDestroy";
  static final String SOCKET_DESTROY_ALL = "socketDestroyAll";
}
